/* vi:ts=4:sw=4
 *
 * VIM - Vi IMproved		by Bram Moolenaar
 *
 * Do ":help uganda"  in Vim to read copying and usage conditions.
 * Do ":help credits" in Vim to see a list of people who contributed.
 */

/*
 * help.c: open a read-only window on the vim_help.txt file
 */

#include "vim.h"
#include "globals.h"
#include "proto.h"
#include "option.h"

	void
do_help(arg)
	char_u		*arg;
{
	char_u	*fnamep;
	FILE	*helpfd;			/* file descriptor of help file */
	int		n;
	WIN		*wp;
	int		num_files;
	char_u	**files_found;
	regexp	*prog;
	int		need_free = FALSE;
	int		attempt;
	char_u	*s, *d;

	/*
	 * If an argument is given, check if there is a match for it
	 * Six tries are done:
	 * 0: ^arg$, reg_ic = FALSE
	 * 1: ^arg   reg_ic = FALSE
	 * 2:  arg   reg_ic = FALSE
	 * 3: ^arg$, reg_ic = TRUE
	 * 4: ^arg   reg_ic = TRUE
	 * 5:  arg   reg_ic = TRUE
	 */
	if (*arg != NUL)
	{
		reg_magic = p_magic;
		for (attempt = 0; attempt < 6; ++attempt)
		{
			if (attempt < 3)
				reg_ic = FALSE;
			else
				reg_ic = TRUE;
			d = IObuff;
			if (attempt != 2 && attempt != 5)
				*d++ = '^';
			/*
			 * Replace "|" with "bar", """ with "quote" and "*" with "star" to
			 * match the name of the tags for these commands.
			 * Replace "*" with ".*" and "?" with "." to match command line
			 * completion.
			 * Insert a backslash before '~', '$' and '.' to avoid their
			 * special meaning.
			 * Replace "^x" by "CTRL-X". Don't do this for "^_" to make
			 * ":help i_^_CTRL-D" work.
			 * If tag starts with ', toss everything after a second '. Fixes
			 * CTRL-] on 'option'. (excluding the '.').
			 */
			if (STRCMP(arg, "*") == 0 || STRCMP(arg, "[*") == 0 ||
													   STRCMP(arg, "]*") == 0)
			{
				if (*arg != '*')
					*d++ = *arg;
				STRCPY(d, "star");
				d += 4;
			}
			else
			{
				for (s = arg; *s; ++s)
				{
					switch (*s)
					{
						case '|':	STRCPY(d, "bar");
									d += 3;
									continue;
						case '\"':	STRCPY(d, "quote");
									d += 5;
									continue;
						case '*':	*d++ = '.';
									break;
									/* "?", ":?" and "?<CR>" are real tags */
						case '?':	if (arg[1] == NUL ||
													 STRCMP(arg, ":?") == 0 ||
													STRCMP(arg, "?<CR>") == 0)
										break;
									*d++ = '.';
									continue;
						case '$':
						case '.':
						case '~':	*d++ = '\\';
									break;
					}
					if (*s < ' ' || (*s == '^' && s[1] && s[1] != '_'))	/* ^x */
					{
						STRCPY(d, "CTRL-");
						d += 5;
						if (*s < ' ')
						{
							*d++ = *s + '@';
							continue;
						}
						++s;
					}
					*d++ = *s;
					if (*s == '\'' && s > arg && *arg == '\'')
						break;
				}
			}
			if (attempt == 0 || attempt == 3)
				*d++ = '$';
			*d = NUL;
			prog = vim_regcomp(IObuff);
			if (prog == NULL)
				return;
			files_found = (char_u **)"";
			num_files = 0;
			n = find_tags(NULL, prog, &num_files, &files_found, TRUE);
			vim_free(prog);
			if (n == FAIL || num_files)
				break;
		}
		if (num_files == 0 || n == FAIL)
		{
			EMSG2("Sorry, no help for %s", arg);
			return;
		}

		/*
		 * If there is at least one match, use the first one.
		 */
		arg = strsave(files_found[0]);
		need_free = TRUE;
		FreeWild(num_files, files_found);
	}

	/*
	 * If there is already a help window open, use that one.
	 */
	if (!curwin->w_buffer->b_help)
	{
		for (wp = firstwin; wp != NULL; wp = wp->w_next)
			if (wp->w_buffer != NULL && wp->w_buffer->b_help)
				break;
		if (wp != NULL && wp->w_buffer->b_nwindows > 0)
			win_enter(wp, TRUE);
		else
		{
			/*
			 * try to open the file specified by the "helpfile" option
			 */
			fnamep = p_hf;
			if ((helpfd = fopen((char *)p_hf, READBIN)) == NULL)
			{
#if defined(MSDOS)
			/*
			 * for MSDOS: try the DOS search path
			 */
				fnamep = searchpath("vim_help.txt");
				if (fnamep == NULL ||
							(helpfd = fopen((char *)fnamep, READBIN)) == NULL)
				{
					smsg((char_u *)"Sorry, help file \"%s\" and \"vim_help.txt\" not found", p_hf);
					goto erret;
				}
#else
				smsg((char_u *)"Sorry, help file \"%s\" not found", p_hf);
				goto erret;
#endif
			}
			fclose(helpfd);

			if (win_split(0, FALSE) == FAIL)
				goto erret;
			
			if (curwin->w_height < p_hh)
				win_setheight((int)p_hh);

#ifdef RIGHTLEFT
			curwin->w_p_rl = 0;				/* help window is left-to-right */

			/* Clear when there are rl windows on the screen. */
			for (wp = firstwin; wp != NULL && !wp->w_p_rl; wp = wp->w_next)
				;
			if (wp != NULL)
				updateScreen(CLEAR);
#endif

			/* open help file (do_ecmd() will set b_help flag, readfile() will
			 * set b_p_ro flag) */
			(void)do_ecmd(0, fnamep, NULL, NULL, TRUE, (linenr_t)0, TRUE);

			/* save the values of the options we change */
			vim_free(help_save_isk);
			help_save_isk = strsave(curbuf->b_p_isk);
			help_save_ts = curbuf->b_p_ts;

			/* accept all chars for keywords, except ' ', '*', '"', '|' */
			set_string_option((char_u *)"isk", -1,
											 (char_u *)"!-~,^*,^|,^\"", TRUE);
			curbuf->b_p_ts = 8;
			check_buf_options(curbuf);
			(void)init_chartab();
		}
	}

	restart_edit = 0;		/* don't want insert mode in help file */

	stuffReadbuff((char_u *)":ta ");
	if (arg != NULL && *arg != NUL)
		stuffReadbuff(arg);
	else
		stuffReadbuff((char_u *)"vim_help.txt");		/* go to the index */
	stuffcharReadbuff('\n');

erret:
	if (need_free)
		vim_free(arg);
}
